from collections.abc import Callable, Iterable, Mapping
from pathlib import Path
from typing import Any, Unpack

import pandas as pd
from rics.misc import get_by_full_name, tname
from rics.types import AnyPath

from ..offline.types import PlaceholderTranslations
from ..translator_typing import AbstractFetcherParams
from ..types import IdType
from ._abstract_fetcher import AbstractFetcher
from .types import FetchInstruction

PandasReadFunction = Callable[[AnyPath], pd.DataFrame]
FormatFn = Callable[[str], str]


class PandasFetcher(AbstractFetcher[str, IdType]):
    """Fetcher implementation using :class:`pandas.DataFrame` as the data format.

    Fetch data from serialized frames. How this is done is determined by the `read_function`. This is typically a Pandas
    function such as :func:`pandas.read_csv` or :func:`pandas.read_parquet`, but any function that accepts a string or
    path `source` as the first argument and returns a data frame can be used.

    .. hint::

       When using **remote file systems**, :attr:`~.AbstractFetcher.sources` are resolved using
       `AbstractFileSystem.glob()`_. If resolution fails, consider overriding the :meth:`find_sources`-method.

    Args:
        read_function: A Pandas `read`-function. String are resolved by :func:`~rics.misc.get_by_full_name`, using
            ``default_module=pandas`` for convenience.
        read_path_format: A string on the form ``protocol://path/to/sources/{}.<ext>``, or a callable to apply
            to a source before passing them to `read_function`.
        read_function_kwargs: Additional keyword arguments for `read_function`.
        **kwargs: See :class:`.AbstractFetcher`.

    See Also:
        The official `Pandas IO documentation <https://pandas.pydata.org/pandas-docs/stable/user_guide/io.html>`_

    .. _AbstractFileSystem.glob(): https://filesystem-spec.readthedocs.io/en/latest/api.html?highlight=glob#fsspec.spec.AbstractFileSystem.glob
    """

    def __init__(
        self,
        read_function: PandasReadFunction | str = "read_csv",
        read_path_format: str | FormatFn = "data/{}.csv",
        read_function_kwargs: Mapping[str, Any] | None = None,
        **kwargs: Unpack[AbstractFetcherParams[str, IdType]],
    ) -> None:
        super().__init__(**kwargs)
        self._read = get_by_full_name(read_function, pd) if isinstance(read_function, str) else read_function
        self._format_source: FormatFn = read_path_format if callable(read_path_format) else read_path_format.format
        self._kwargs = read_function_kwargs or {}

        self._source_paths: dict[str, str] = {}

    def read(self, source_path: AnyPath) -> pd.DataFrame:
        """Read a ``DataFrame`` from a source path.

        Args:
            source_path: Path to serialized ``DataFrame``.

        Returns:
            A deserialized ``DataFrame``.
        """
        return self._read(source_path, **self._kwargs).convert_dtypes()

    def format_source(self, source: str) -> str:
        """Get the path for `source`."""
        return self._format_source(source)

    def find_sources(self, task_id: int = -1) -> dict[str, str]:
        """Resolve sources and their associated paths.

        Args:
            task_id: Used for logging.

        Sources are resolved in three steps:

        1. Create glob pattern by calling :meth:`format_source` with ``source='*'``.
        2. Glob files using `AbstractFileSystem.glob()`_ (requires ``fsspec``) or :meth:`Path.glob() <pathlib.Path.glob>`.
        3. Strip the directory and file suffix from the globbed paths to create source names.

        .. _AbstractFileSystem.glob(): https://filesystem-spec.readthedocs.io/en/latest/api.html?highlight=glob#fsspec.spec.AbstractFileSystem.glob

        Returns:
            A dict ``{source: path}``.
        """
        pattern = self.format_source("*")

        try:
            source_paths = self._find_sources_fsspec(pattern)
        except ModuleNotFoundError as e:
            self.logger.debug(f"Falling back to 'pathlib.Path': {e!r}")
            source_paths = self._find_sources_pathlib(pattern)

        extra = {"task_id": task_id, "pattern": pattern}
        if source_paths:
            self.logger.debug(f"Path {pattern=} matched {len(source_paths)} files: {source_paths}", extra=extra)
            return source_paths
        else:
            self.logger.warning(f"Path {pattern=} did not match any files.", extra=extra)
            return {}

    def _find_sources_fsspec(self, pattern: str) -> dict[str, str]:
        from fsspec.core import url_to_fs  # type: ignore

        fs, _ = url_to_fs(pattern, **self._kwargs.get("storage_options", {}))

        source_paths = self._make_source_paths(fs.glob(pattern))

        protocol, separator, _ = pattern.partition("://")
        if separator:
            source_paths = {source: protocol + separator + path for source, path in source_paths.items()}

        return source_paths

    def _find_sources_pathlib(self, pattern: str) -> dict[str, str]:
        path = Path(pattern)
        iterator = path.parent.glob(path.name)
        return self._make_source_paths(iterator)

    @staticmethod
    def _make_source_paths(iterator: Iterable[Path]) -> dict[str, str]:
        source_paths = {}
        for path in map(Path, iterator):
            source = path.with_suffix("").name
            source_paths[source] = str(path)
        return source_paths

    def _initialize_sources(self, task_id: int) -> dict[str, list[str]]:
        self._source_paths = self.find_sources(task_id)
        return {source: self.read(path).columns.tolist() for source, path in self._source_paths.items()}

    def fetch_translations(self, instr: FetchInstruction[str, IdType]) -> PlaceholderTranslations[str]:
        return PlaceholderTranslations.make(
            instr.source,
            self.read(self._source_paths[instr.source]),
        )

    def __repr__(self) -> str:
        read_path_format = self.format_source("{}")
        return f"{tname(self)}(read_function={tname(self._read)}, {read_path_format=})"



"""Integration for insertion and extraction of IDs and translations to and from various data structures.

User-defined integrations
-------------------------
The purpose of creating new integrations is typically to enable translation of a new data type.
To get started, inherit from :class:`DataStructureIO` or copy an
:class:`existing <id_translation.dio.integration.polars>` integration. Don't forget to
:meth:`register <.DataStructureIO.register>` the implementation, or the :class:`.Translator` won't be able to find it.

Automatic integration discovery
-------------------------------
You may add an entrypoint in the ``{entrypoint_group!r}`` entrypoint group to
automatically register custom implementations (as opposed to calling :meth:`.DataStructureIO.register` manually). The
snippet below shows how the :mod:`bundled <.integration>` integrations are registered using project entrypoints.

.. code-block:: toml
   :caption: Entrypoints in ``pyproject.toml`` in the
        https://github.com/rsundqvist/id-translation/blob/v0.14.0/pyproject.toml#L45-L48 project.

   [project.entry-points."id_translation.dio"]
   # The name (e.g. 'dask_io') is not important, but should be unique.
   dask_io = "id_translation.dio.integration.dask:DaskIO"
   polars_io = "id_translation.dio.integration.polars:PolarsIO"

The :func:`loader <id_translation.dio.load_integrations>` will skip the integration if calling
:class:`EntryPoint.load() <importlib.metadata.EntryPoint>` raises an :py:class:`ImportError`.

.. autodata:: id_translation.dio::ENTRYPOINT_GROUP
"""

# The autodata is a workaround for ENTRYPOINT_GROUP
#   https://github.com/sphinx-doc/sphinx/issues/6495#issuecomment-1058033697
#   https://github.com/sphinx-doc/sphinx/issues/12020

from ._data_structure_io import DataStructureIO
from ._resolve import (
    ENTRYPOINT_GROUP,
    get_resolution_order,
    is_registered,
    load_integrations,
    register_io,
    resolve_io,
)

__all__ = [
    "ENTRYPOINT_GROUP",
    "DataStructureIO",
    "get_resolution_order",
    "is_registered",
    "load_integrations",
    "register_io",
    "resolve_io",
]

load_integrations()
__doc__ = __doc__.format(entrypoint_group=ENTRYPOINT_GROUP)



from typing import Any

from rics.misc import get_by_full_name

from id_translation.fetching import CacheAccess
from id_translation.types import IdType, SourceType


def default_cache_access_factory(clazz: str, config: dict[str, Any]) -> CacheAccess[SourceType, IdType]:
    """Create a :class:`.CacheAccess` from config."""
    cls = get_by_full_name(
        clazz,
        subclass_of=CacheAccess,  # type: ignore[type-abstract]  # https://github.com/python/mypy/issues/4717
    )
    return cls(**config)



from typing import Any

from rics.misc import get_by_full_name

from id_translation.fetching import AbstractFetcher
from id_translation.types import IdType, SourceType


def default_fetcher_factory(clazz: str, config: dict[str, Any]) -> AbstractFetcher[SourceType, IdType]:
    """Create an :class:`.AbstractFetcher` from config."""
    from id_translation import fetching as default_module

    cls = get_by_full_name(
        clazz,
        default_module,
        subclass_of=AbstractFetcher,  # type: ignore[type-abstract]  # https://github.com/python/mypy/issues/4717
    )
    return cls(**config)



from collections.abc import Sequence
from typing import Any, TypeVar
from uuid import UUID

import numpy as np

IdCollectionT = TypeVar("IdCollectionT", bound=Sequence)  # type: ignore[type-arg] # TODO: Need Higher-Kinded TypeVars


def cast_many(ids: IdCollectionT) -> IdCollectionT:
    """Cast `ids` to UUIDs."""
    uuids = map(UUID, ids)
    if isinstance(ids, np.ndarray):
        return np.fromiter(uuids, dtype=UUID)
    else:
        return type(ids)(uuids)  # type: ignore[call-arg]


def try_cast_many(ids: IdCollectionT) -> IdCollectionT:
    """Try casting `ids` to UUIDs, falling back to the original input."""
    if len(ids) == 0:
        return ids

    try:
        return cast_many(ids)
    except (ValueError, AttributeError, TypeError):
        return ids


def try_cast_one(idx: Any) -> Any:
    """Try casting `idx` to a UUID, falling back to the original input."""
    if isinstance(idx, UUID):
        return idx

    try:
        return UUID(idx)
    except (ValueError, AttributeError, TypeError):
        return idx



from typing import Any

from rics.misc import get_by_full_name

from id_translation.transform.types import Transformer
from id_translation.types import IdType


def default_transformer_factory(clazz: str, config: dict[str, Any]) -> Transformer[IdType]:
    """Create a :class:`.Transformer` from config."""
    from id_translation import transform as default_module

    cls = get_by_full_name(
        clazz,
        default_module,
        subclass_of=Transformer,  # type: ignore[type-abstract]  # https://github.com/python/mypy/issues/4717
    )
    return cls(**config)



import logging
import warnings
from collections import defaultdict
from collections.abc import Iterable, Sequence
from time import perf_counter
from typing import TYPE_CHECKING, Any, get_args

from numpy import isnan, unique
from rics.misc import tname
from rics.strings import format_seconds as fmt_sec

from .. import _uuid_utils
from ..exceptions import TooManyFailedTranslationsError
from ..mapping.types import UserOverrideFunction
from ..offline import Format, TranslationMap
from ..settings import logging as settings
from ..types import IdType, IdTypes, Names, NameToSource, NameType, NameTypes, SourceType, Translatable
from ..utils.logging import cast_unsafe
from ._map import MappingTask

LOGGER = logging.getLogger("id_translation.Translator.translate")

NUM_SAMPLE_IDS = 10

if TYPE_CHECKING:
    from .._translator import Translator


class TranslationTask(MappingTask[NameType, SourceType, IdType]):
    """Ephemeral class for performing a single translation task on a `translatable`."""

    def __init__(
        self,
        caller: "Translator[NameType, SourceType, IdType]",
        translatable: Translatable[NameType, IdType],
        fmt: Format,
        names: NameTypes[NameType] | NameToSource[NameType, SourceType] | None = None,
        *,
        ignore_names: Names[NameType] = None,
        override_function: UserOverrideFunction[NameType, SourceType, None] | None = None,
        copy: bool = True,
        max_fails: float = 1.0,
        reverse: bool = False,
        enable_uuid_heuristics: bool = False,
    ) -> None:
        super().__init__(
            caller,
            translatable,
            names,
            ignore_names=ignore_names,
            override_function=override_function,
        )

        if not (0.0 <= max_fails <= 1):
            raise ValueError(f"Argument {max_fails=} is not a valid fraction")

        self.fmt = fmt
        self.copy = copy
        self.max_fails = max_fails
        self.reverse = reverse
        self.parent: Translatable[NameType, IdType] | None = None

        self.enable_uuid_heuristics = enable_uuid_heuristics

        self.key_event_level = settings.TRANSLATE_ONLINE if self.caller.online else settings.TRANSLATE_OFFLINE

        self._names_without_ids: set[NameType] = set()

    @property
    def io_names(self) -> list[NameType]:
        """Names for which IDs should be extracted from the `translatable`."""
        # Preserve input order for names, if given. These names may be repeated.
        if self.names_from_user is None:
            names = self.names_to_translate
        else:
            names = self.names_from_user

        return [n for n in names if n not in self._names_without_ids]

    def extract_ids(self) -> dict[SourceType, set[IdType]]:
        """Extract IDs to fetch from the translatable."""
        start = perf_counter()
        name_to_source = self.name_to_source
        source_to_ids: dict[SourceType, set[IdType]] = defaultdict(set)

        float_names: list[NameType] = []
        num_coerced = 0
        ids: Sequence[IdType]
        for name, ids in self._extract_ids().items():
            if len(ids) == 0:
                continue

            all_ids = source_to_ids[name_to_source[name]]

            was_coerced: bool = False
            if isinstance(ids[0], float):
                # Float IDs aren't officially supported, but is common when using Pandas since int types cannot be NaN.
                # This is sometimes a problem for the built-in set (see https://github.com/numpy/numpy/issues/9358), and
                # for several database drivers.

                try:
                    ids, n_new = self._coerce_float_to_int(ids)  # noqa: PLW2901
                    num_coerced += n_new  # Somewhat inaccurate; includes repeat IDs from other names
                    float_names.append(name)
                    was_coerced = True
                except TypeError:
                    pass

            if not was_coerced and self.enable_uuid_heuristics:
                ids = _uuid_utils.try_cast_many(ids)  # noqa: PLW2901

            all_ids.update(ids)

        if num_coerced > 100:  # pragma: no cover  # noqa: PLR2004
            types = f"({', '.join(t.__name__ for t in get_args(IdTypes))})"
            warnings.warn(
                f"To ensure proper fetcher operation, {num_coerced} float-type IDs have been coerced to int. "
                f"Enforcing supported data types {types} for IDs in your {self.type_name}-data may improve performance."
                f" Affected names ({len(float_names)}): {float_names}.",
                stacklevel=3,
            )

        execution_time = perf_counter() - start
        self.add_timing("extract", execution_time)

        return source_to_ids

    @classmethod
    def _coerce_float_to_int(cls, ids: Sequence[float]) -> tuple[Iterable[int], int]:
        arr = unique(ids)
        keep_mask = ~isnan(arr, casting="no")
        arr = arr[keep_mask]
        arr = arr.astype(int, copy=False)
        return arr, keep_mask.sum()

    def log_key_event_enter(self) -> None:
        """Emits the ``TRANSLATOR:TRANSLATE.ENTER`` message."""
        if not LOGGER.isEnabledFor(self.key_event_level.enter):
            return

        type_name = self.type_name
        names = self.names_from_user
        ignore_names = self.ignore_names

        name_info = f"Derive based on type={type_name}" if names is None else repr(names)
        if ignore_names is not None:
            name_info += f", excluding those given by {ignore_names=}"

        LOGGER.log(
            level=self.key_event_level.enter,
            msg=f"Begin translation of {type_name}-type data. Names to translate: {name_info}.",
            extra=dict(
                task_id=self.task_id,
                event_key="TRANSLATOR.TRANSLATE",
                event_stage="ENTER",
                event_title="TRANSLATOR.TRANSLATE.ENTER",
                sources=self.caller.sources,
                online=self.caller.online,
                # Task-specific
                translatable_type=self.full_type_name,
                names=names,
                ignore_names=tname(ignore_names, prefix_classname=True) if callable(ignore_names) else ignore_names,
                copy=self.copy,
                reverse=self.reverse,
            ),
        )

    def log_key_event_exit(self) -> None:
        """Emits the ``TRANSLATOR:TRANSLATE.EXIT`` key event message."""
        if not LOGGER.isEnabledFor(self.key_event_level.exit):
            return

        copy = self.copy
        n2s = self.name_to_source
        execution_time = perf_counter() - self._start

        LOGGER.log(
            level=self.key_event_level.exit,
            msg=(
                f"Finished translation of {len(n2s)} names in {self.type_name}-type data in "
                f"{fmt_sec(execution_time)}, using name-to-source mapping: {n2s}."
            ),
            extra=dict(
                task_id=self.task_id,
                event_key="TRANSLATOR.TRANSLATE",
                event_stage="EXIT",
                event_title="TRANSLATOR.TRANSLATE.EXIT",
                execution_time=execution_time,
                duration_ms=self.get_timings(),
                # Task-specific
                translatable_type=self.full_type_name,
                copy=copy,
                reverse=self.reverse,
                name_to_source=n2s,
            ),
        )

    def insert(
        self, translation_map: TranslationMap[NameType, SourceType, IdType]
    ) -> Translatable[NameType, str] | None:
        """Insert translated IDs into the `translatable`, based on data retrieved by the fetcher."""
        start = perf_counter()
        copy = self.copy

        translation_map.reverse_mode = self.reverse
        try:
            result = self.io.insert(
                self.translatable,
                names=self.io_names,
                tmap=translation_map,
                copy=copy,
            )
        finally:
            translation_map.reverse_mode = False

        execution_time = perf_counter() - start
        self.add_timing("insert", execution_time)
        return result if copy else None

    def verify(self, tmap: TranslationMap[NameType, SourceType, IdType]) -> None:
        """Verify translations.

        Performs translation with pre-defined formats, counting the number of IDs which are (and aren't) known to the
        translation map.
        """
        start = perf_counter()

        if not (self.max_fails < 1.0 or LOGGER.isEnabledFor(logging.DEBUG)):
            return

        name_to_ids = self._name_to_ids_in_order()
        translations = tmap.to_translations()

        for name, ids in name_to_ids.items():
            source = tmap.name_to_source[name]
            magic_dict = translations[source]

            if self.reverse == tmap.reverse_mode:
                is_known = magic_dict.real_contains
            else:
                is_known = {*magic_dict.real.values()}.__contains__

            is_missing = [not is_known(idx) for idx in ids]
            n_untranslated = sum(is_missing)
            if n_untranslated == 0:
                continue
            n_total = len(ids)
            f_untranslated = n_untranslated / n_total

            sample_ids = self._get_untranslated_ids(name_to_ids[name], is_missing_mask=is_missing)

            extra = {"name_of_ids": name, "source": source, "sample_ids": cast_unsafe(sample_ids)}
            message = (
                f"Failed to translate {n_untranslated}/{n_total} ({f_untranslated:.1%}{{reason}}) of IDs "
                f"for {name=} using source={source!r}. Sample IDs: {sample_ids}."
            )

            if f_untranslated > self.max_fails:
                message = message.format(reason=f" > max_fails={self.max_fails:.1%}")
                LOGGER.error(message, extra=extra)
                raise TooManyFailedTranslationsError(message)
            else:
                message = message.format(reason=", above limit; DEBUG logging is enabled")
                LOGGER.debug(message, extra=extra)

        execution_time = perf_counter() - start
        self.add_timing("verify", execution_time)

    def _name_to_ids_in_order(self) -> dict[NameType, Sequence[Any]]:
        name_to_ids: dict[NameType, Sequence[Any]] = self._extract_ids()
        if self.enable_uuid_heuristics:
            name_to_ids = {name: _uuid_utils.try_cast_many(name_to_ids[name]) for name in name_to_ids}
        return name_to_ids

    def _extract_ids(self) -> dict[NameType, Sequence[IdType]]:
        name_to_ids: dict[NameType, Sequence[IdType]] = self.io.extract(self.translatable, self.io_names)

        name_to_ids = {name: ids for name, ids in name_to_ids.items() if len(ids) > 0}
        if not self._names_without_ids:
            self._names_without_ids = set(self.io_names).difference(name_to_ids)
        return name_to_ids

    @staticmethod
    def _get_untranslated_ids(ids: Sequence[IdType], *, is_missing_mask: Sequence[bool]) -> list[IdType]:
        seen = set()
        retval = []

        for idx, is_missing in zip(ids, is_missing_mask, strict=True):
            if not is_missing or idx in seen:
                continue

            seen.add(idx)
            retval.append(idx)
            if len(retval) == NUM_SAMPLE_IDS:
                break

        return retval
